from ..framework.loggable import Loggable
from .api_client import APIClient


class APIResource(Loggable):
    EMPTY_PATH_PARAMS = {}

    def __init__(self, base_endpoint_template: str, api_client: APIClient):
        super().__init__()

        self.base_endpoint_template = base_endpoint_template
        self.api_client = api_client

    def find(self, path_params: dict, headers=None):
        endpoint = self.__generate_endpoint(path_params)

        response = self.api_client.get(endpoint=endpoint, headers=headers)

        return response

    def list(self, path_params: dict, params: dict = None, headers=None):
        endpoint = self.__generate_endpoint(path_params)
        response = self.api_client.get(endpoint, params=params, headers=headers)

        return response

    def create(
        self,
        path_params: dict = None,
        payload=None,
        headers=None,
        expected_successful_status_code=201,
    ):
        endpoint = self.__generate_endpoint(path_params)

        response = self.api_client.post(
            endpoint, payload, headers, expected_successful_status_code
        )

        return response

    def update(
        self,
        path_params: dict,
        payload,
        headers=None,
        expected_successful_status_code=200,
    ):
        endpoint = self.__generate_endpoint(path_params)

        response = self.api_client.patch(
            endpoint, payload, headers, expected_successful_status_code
        )

        return response

    def full_update(
        self,
        path_params: dict,
        payload,
        headers=None,
        expected_successful_status_code=200,
    ):
        endpoint = self.__generate_endpoint(path_params)

        response = self.api_client.put(
            endpoint, payload, headers, expected_successful_status_code
        )

        return response

    def delete(self, path_params: dict, headers=None):

        endpoint = self.__generate_endpoint(path_params)

        self.api_client.delete(endpoint, headers)

    def __generate_endpoint(self, path_params):
        url_parts = []

        parts = self.base_endpoint_template.split("/")

        for part in parts:
            if part.startswith("{") and part.endswith("}"):
                param_name = part[1:-1]

                if param_name in path_params:
                    url_parts.append(str(path_params[param_name]))
            else:
                url_parts.append(part)

        url = "/".join(url_parts)

        return url



import os

from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait


def wait_for_element(driver, condition, timeout=10) -> WebElement:
    try:
        return WebDriverWait(driver, timeout).until(condition)
    except Exception as e:
        screenshot_path = os.path.join(os.getcwd(), "screenshots/wait_exception.png")

        driver.save_screenshot(screenshot_path)

        raise e


def wait_for_element_by_xpath(driver, xpath: str, timeout=10) -> WebElement:
    return wait_for_element(
        driver, EC.visibility_of_element_located((By.XPATH, xpath)), timeout
    )


def wait_for_element_by_selector(driver, css_selector: str, timeout=10) -> WebElement:
    return wait_for_element(
        driver,
        EC.visibility_of_element_located((By.CSS_SELECTOR, css_selector)),
        timeout,
    )


def wait_for_element_to_disappear_by_xpath(
    driver, xpath: str, timeout=10
) -> WebElement:
    return WebDriverWait(driver, timeout).until(
        EC.invisibility_of_element_located((By.XPATH, xpath))
    )


def wait_for_element_to_click_by_xpath(driver, xpath: str, timeout=10):
    return wait_for_element(
        driver, EC.element_to_be_clickable((By.XPATH, xpath)), timeout
    )



from abc import ABC, abstractmethod

from .loggable import Loggable


class Module(Loggable, ABC):
    app = None

    def __init__(self, app) -> None:
        super().__init__()

        self.app = app

        self.bootstrap()

    @abstractmethod
    def bootstrap(self):
        pass



import datetime
import time
from functools import wraps
from typing import Callable, TypeVar

T = TypeVar("T")


def with_timeout(timeout: int = 10):
    def decorator(func: Callable):
        @wraps(func)
        def wrapper(self, *args, **kwargs):

            number_of_args = len(args)
            timeout_arg = args[number_of_args - 1]
            timeout_arg_existed = f"{timeout_arg}".isnumeric()

            if "timeout" in kwargs:
                remaining_timeout = kwargs["timeout"]
            elif timeout_arg_existed:
                remaining_timeout = timeout_arg
            else:
                remaining_timeout = timeout

            start_time = datetime.datetime.now()

            try:
                return func(self, *args, **kwargs)
            except Exception as e:
                end_time = datetime.datetime.now()
                time_difference = (end_time - start_time).total_seconds()
                new_remaining_timeout = remaining_timeout - time_difference - 1

                if new_remaining_timeout > 0:
                    kwargs["timeout"] = new_remaining_timeout

                    time.sleep(1)

                    return wrapper(self, *args, **kwargs)
                else:
                    raise e

        return wrapper

    return decorator



import array

from transstellar.framework.loggable import Loggable

from .main_config import MainConfig
from .route import Route


class Router(Loggable):
    app: None
    routes: dict = {}

    def __init__(self, app):
        super().__init__()

        self.app = app

    def register_routes(self, routes: array):
        if routes is None:
            return

        for route in routes:
            self.register_route(route)

    def register_route(self, route: Route):
        if not isinstance(route, Route):
            raise TypeError("route is not a Route instance")

        self.routes[route.route_key] = route

    def get_route(self, route_key: str):
        route = self.routes.get(route_key)

        if route is None:
            raise LookupError(f'Route "{route_key}" not found')

        return route

    def get_page(self, route_key: str):
        route = self.get_route(route_key)

        page = route.get_page(self.app)

        return page

    def build_page(self, page_class):
        return page_class.create_element(self)

    def go_to_url(self, url: str):
        if url != self.app.driver.current_url:
            self.logger.info("Go to url: %s", url)
            self.app.driver.get(url)

    def go_to(self, route_key: str, path_params: dict = None):
        route = self.get_route(route_key)
        main_config = self.app.get(MainConfig)

        base_url = main_config.get_app_url()
        path = route.path

        if path_params:
            path = path.format(**path_params)

        url = f"{base_url}{path}"

        self.go_to_url(url)

        page = route.get_page(self.app)

        return page



import logging


class Logger:
    class_name: str

    def __init__(self, class_name) -> None:
        self.class_name = class_name

    def configure(self, file_format, filename, level):
        logging.basicConfig(
            format=file_format,
            filename=filename,
            level=level,
        )

    def debug(self, msg, *args, **kwargs):
        logging.debug(self.__get_message(msg), *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        logging.info(self.__get_message(msg), *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        logging.debug(self.__get_message(msg), *args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        logging.debug(self.__get_message(msg), *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        logging.debug(self.__get_message(msg), *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        logging.debug(self.__get_message(msg), *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        logging.debug(self.__get_message(msg), *args, **kwargs)

    def fatal(self, msg, *args, **kwargs):
        logging.debug(self.__get_message(msg), *args, **kwargs)

    def __get_message(self, msg: str):
        return f"[{self.class_name}] {msg}"



import time

import pytest

from .application import Application
from .logger import Logger


class BaseTest:
    app: Application
    logger: Logger

    @pytest.fixture(autouse=True)
    def setup_method_base_test(self, app: Application):
        self.app = app
        self.logger = Logger(self.__class__.__name__)

    def sleep(self, seconds: float):
        self.logger.debug(f"sleep: {seconds} seconds")

        time.sleep(seconds)



import os
from typing import Type
from urllib.parse import ParseResult, urlparse

from injector import Injector
from pytest import FixtureRequest
from selenium.webdriver import ChromeOptions, Remote
from selenium.webdriver.remote.webdriver import WebDriver

from transstellar.framework.loggable import Loggable

from .main_config import MainConfig
from .module import Module
from .router import Router


# pylint: disable-next=R0902
class Application(Loggable):
    container: Injector
    testrun_uid: str
    request: FixtureRequest
    driver: WebDriver
    options: any
    e2e_enabled: bool = False
    closed: bool = False
    current_user = None

    def __init__(self, params: dict):
        super().__init__()

        self.logger.info("Creating application")

        options = params.get("options")
        self.request = params.get("request")
        self.driver = params.get("driver")
        self.testrun_uid = params.get("testrun_uid")
        self.container = Injector()
        self.router = Router(self)
        self.main_config = self.get(MainConfig)

        routes = params.get("routes")

        if options:
            self.options = options
        else:
            self.options = {}

        self.__configure_log__()
        self.__remove_existing_screenshots__()
        self.register_routes(routes)

    def init_e2e(self):
        if self.e2e_enabled:
            return

        self.e2e_enabled = True

        self.driver = self.__init_driver__()

        self.closed = False

    def is_e2e_enabled(self):
        return self.e2e_enabled

    def get(self, key: any):
        return self.container.get(key)

    def register_module(self, module_class: Type[Module]):
        module = module_class(self)
        self.container.binder.bind(module_class, module)

    def register_routes(self, routes: dict):
        self.router.register_routes(routes)

    def get_page(self, route_key: str):
        self.guard_e2e()

        return self.router.get_page(route_key)

    def build_page(self, page_class):
        return self.router.build_page(page_class)

    def go_to_url(self, url: str):
        self.router.go_to_url(url)

    def go_to(self, route_key: str, path_params: dict = None):
        self.guard_e2e()

        return self.router.go_to(route_key, path_params)

    def guard_e2e(self):
        if not self.e2e_enabled:
            raise AttributeError("e2e is not enabled")

    def close(self):
        if self.closed:
            return

        self.logger.info("Closing application")

        if self.is_e2e_enabled():
            self.driver.quit()
            self.logger.info("Driver closed")

        self.e2e_enabled = False
        self.closed = True

        self.logger.info("Application closed")

    def set_current_user(self, user):
        if not self.is_e2e_enabled():
            raise RuntimeError("Only E2E test supports setting current user")

        self.current_user = user

    def get_current_user(self):
        return self.current_user

    def is_logged_in(self):
        return self.current_user is not None

    def get_current_url(self) -> ParseResult:
        return urlparse(self.driver.current_url)

    def __configure_log__(self):
        worker_id = os.environ.get("PYTEST_XDIST_WORKER")
        if worker_id is not None:
            with open(file=f"logs/pytest_{worker_id}.log", mode="w", encoding="utf-8"):
                pass

            self.logger.configure(
                file_format=self.request.config.getini("log_file_format"),
                filename=f"logs/pytest_{worker_id}.log",
                level=self.request.config.getini("log_file_level"),
            )

    def __remove_existing_screenshots__(self):
        screenshots_dir = os.path.join(os.getcwd(), "screenshots")

        if not os.path.exists(screenshots_dir):
            os.makedirs(screenshots_dir)
        else:
            for file_name in os.listdir(screenshots_dir):
                file_path = os.path.join(screenshots_dir, file_name)
                if os.path.isfile(file_path):
                    os.remove(file_path)

    def __init_driver__(self) -> WebDriver:
        self.logger.info("Initializing driver")
        selenium_cmd_executor = os.environ.get(
            "SELENIUM_CMD_EXECUTOR", "http://selenium:4444/wd/hub"
        )
        implicitly_wait_time = self.options.get("implicitly_wait_time", 10)
        options = ChromeOptions()
        options.add_argument("--no-sandbox")
        options.add_argument("--disable-dev-shm-usage")
        driver = Remote(command_executor=selenium_cmd_executor, options=options)
        driver.implicitly_wait(implicitly_wait_time)
        self.logger.info("Driver initialized")

        return driver



__all__ = [
    "Application",
    "ApplicationBootstrapper",
    "BaseApiTest",
    "BasePage",
    "BaseTest",
    "BaseUITest",
    "ConfigService",
    "Element",
    "handle_error",
    "handle_ui_error",
    "handle_page_error",
    "Loggable",
    "Logger",
    "MainConfig",
    "Module",
    "Route",
    "Router",
]

from .application import Application
from .application_bootstrapper import ApplicationBootstrapper
from .base_api_test import BaseApiTest
from .base_page import BasePage
from .base_test import BaseTest
from .base_ui_test import BaseUITest
from .config_service import ConfigService
from .decorators.handle_error import handle_error
from .decorators.handle_ui_error import handle_ui_error
from .element import Element
from .handle_page_error import handle_page_error
from .loggable import Loggable
from .logger import Logger
from .main_config import MainConfig
from .module import Module
from .route import Route
from .router import Router



import http
import json

import requests

from transstellar.framework.loggable import Loggable

from .errors import ClientError, ServerError, UnauthorizedError


class APIClient(Loggable):
    token = ""
    headers = {"content-type": "application/json", "accept": "application/json"}

    def __init__(self, base_url: str, options=None):
        super().__init__()

        self.base_url = base_url

        if options is None:
            options = {}

        if options.get("debug") is True:

            def httpclient_log(*args):
                self.logger.debug(" ".join(args))

            # mask the print() built-in in the http.client module to use
            # logging instead
            http.client.print = httpclient_log
            # enable debugging
            http.client.HTTPConnection.debuglevel = 1

    def as_token(self, token):
        self.token = token

        return self

    def get(self, endpoint, params=None, headers=None):

        url = f"{self.base_url}/{endpoint}"

        return self.__get(url, params, headers)

    def post(
        self, endpoint, payload=None, headers=None, expected_successful_status_code=201
    ):
        url = f"{self.base_url}/{endpoint}"
        return self.__post(url, payload, headers, expected_successful_status_code)

    def patch(
        self, endpoint, payload, headers=None, expected_successful_status_code=200
    ):
        url = f"{self.base_url}/{endpoint}"
        return self.__patch(url, payload, headers, expected_successful_status_code)

    def put(self, endpoint, payload, headers=None, expected_successful_status_code=200):
        url = f"{self.base_url}/{endpoint}"
        return self.__put(url, payload, headers, expected_successful_status_code)

    def delete(self, endpoint, headers=None):
        url = f"{self.base_url}/{endpoint}"

        return self.__delete(url, headers)

    def __get(self, url, params, headers):
        headers = self.__get_headers(headers)

        response = requests.get(url, params=params, headers=headers, timeout=10)

        return self.__handle_response(response)

    def __post(self, url, payload, headers=None, expected_successful_status_code=201):
        headers = self.__get_headers(headers)
        response = requests.post(url, json=payload, headers=headers, timeout=10)

        return self.__handle_response(response, expected_successful_status_code)

    def __patch(self, url, payload, headers=None, expected_successful_status_code=200):
        headers = self.__get_headers(headers)
        response = requests.patch(url, json=payload, headers=headers, timeout=10)

        return self.__handle_response(response, expected_successful_status_code)

    def __put(self, url, payload, headers=None, expected_successful_status_code=200):
        headers = self.__get_headers(headers)
        response = requests.put(url, json=payload, headers=headers, timeout=10)

        return self.__handle_response(response, expected_successful_status_code)

    def __delete(self, url, headers=None):
        headers = self.__get_headers(headers)
        response = requests.delete(url, headers=headers, timeout=10)

        return self.__handle_response(response, expected_successful_status_code=204)

    def __get_headers(self, headers):
        if headers is None:
            headers = {}
        return {**self.headers, **headers, "Authorization": f"Bearer {self.token}"}

    def __handle_response(
        self, response: requests.models.Response, expected_successful_status_code=200
    ):
        status_code = response.status_code
        response_json = None

        if response.text:
            response_json = json.loads(response.text)

        if 200 <= status_code < 300:
            if status_code != expected_successful_status_code:
                raise RuntimeError(
                    # pylint: disable-next=C0301
                    "Response status code ({status_code}) is not as expected: {expected_successful_status_code}"
                )
            return response_json

        if 400 <= status_code < 500:
            error_message = response_json.get("message", "Unknown client error")

            if status_code == 400:
                raise ClientError(f"Client error: {error_message}")

            if status_code == 401:
                raise UnauthorizedError("Unauthorized: Check your credentials")

            error_message = response_json.get("message", "Unknown client error")
            raise ClientError(f"Client error: HTTP {status_code}. {error_message}")

        if status_code >= 500:
            error_message = response_json.get("message", "Internal Server Error")
            raise ServerError(f"Server error: HTTP {status_code}. {error_message}")

        raise ServerError(f"Unknown server error: HTTP {status_code}. {error_message}")



import inspect


def handle_ui_error(screenshot_file_name=""):
    def decorator(cls):
        for name, method in vars(cls).items():
            if (
                callable(method)
                and not inspect.isclass(method)
                and not name.startswith("__")
                # NOTE: can not support method with fixture decorator
                and not hasattr(method, "_pytestfixturefunction")
            ):
                setattr(
                    cls,
                    name,
                    handle_ui_errors(method, screenshot_file_name, cls.__name__),
                )

        return cls

    return decorator


def handle_ui_errors(method, screenshot_file_name, class_name):
    def wrapper(*args, **kwargs):
        try:
            result = method(*args, **kwargs)
            return result
        except Exception as e:
            file_name = f"{class_name}_{method.__name__}_error.png"

            if screenshot_file_name:
                file_name = screenshot_file_name
            test_instance = args[0]
            test_instance.screenshot(file_name)

            raise e

    return wrapper



__all__ = [
    "A",
    "Body",
    "Button",
    "Div",
    "H1",
    "H2",
    "Header",
    "Image",
    "Input",
    "Li",
    "Span",
    "Ul",
]

from .a import A
from .body import Body
from .button import Button
from .div import Div
from .h1 import H1
from .h2 import H2
from .header import Header
from .image import Image
from .input import Input
from .li import Li
from .span import Span
from .ul import Ul
